__all__ = ["Spotify", "SpotifyException", "SpotifyAuthException"]

import os
import webbrowser
from pprint import pprint
from typing import Optional, Union

import requests
from dotenv import load_dotenv

from yutipy.base_clients import BaseAuthClient, BaseClient
from yutipy.exceptions import (
    InvalidValueException,
    SpotifyAuthException,
    SpotifyException,
)
from yutipy.logger import logger
from yutipy.lrclib import LrcLib
from yutipy.models import MusicInfo, UserPlaying
from yutipy.utils.helpers import (
    are_strings_similar,
    guess_album_type,
    is_valid_string,
    separate_artists,
)

load_dotenv()

SPOTIFY_CLIENT_ID = os.getenv("SPOTIFY_CLIENT_ID")
SPOTIFY_CLIENT_SECRET = os.getenv("SPOTIFY_CLIENT_SECRET")
SPOTIFY_REDIRECT_URI = os.getenv("SPOTIFY_REDIRECT_URI")


class Spotify(BaseClient):
    """
    A class to interact with the Spotify API. It uses "Client Credentials" grant type (or flow).

    This class reads the ``SPOTIFY_CLIENT_ID`` and ``SPOTIFY_CLIENT_SECRET`` from environment variables or the ``.env`` file by default.
    Alternatively, you can manually provide these values when creating an object.
    """

    def __init__(
        self,
        client_id: str = None,
        client_secret: str = None,
        defer_load: bool = False,
        fetch_lyrics: bool = True,
    ) -> None:
        """
        Parameters
        ----------
        client_id : str, optional
            The Client ID for the Spotify API. Defaults to ``SPOTIFY_CLIENT_ID`` from environment variable or the ``.env`` file.
        client_secret : str, optional
            The Client secret for the Spotify API. Defaults to ``SPOTIFY_CLIENT_SECRET`` from environment variable or the ``.env`` file.
        defer_load : bool, optional
            Whether to defer loading the access token during initialization, by default ``False``
        fetch_lyrics : bool, optional
            Whether to fetch lyrics (using `LRCLIB <https://lrclib.net>`__) if the music platform does not provide lyrics (default is True).
        """
        self.client_id = client_id or SPOTIFY_CLIENT_ID
        self.client_secret = client_secret or SPOTIFY_CLIENT_SECRET
        self.fetch_lyrics = fetch_lyrics

        if not self.client_id:
            raise SpotifyException(
                "Client ID was not found. Set it in environment variable or directly pass it when creating object."
            )

        if not self.client_secret:
            raise SpotifyException(
                "Client Secret was not found. Set it in environment variable or directly pass it when creating object."
            )

        super().__init__(
            service_name="Spotify",
            access_token_url="https://accounts.spotify.com/api/token",
            client_id=self.client_id,
            client_secret=self.client_secret,
            defer_load=defer_load,
        )

        self.__api_url = "https://api.spotify.com/v1"

    def search(
        self,
        artist: str,
        song: str,
        limit: int = 10,
        normalize_non_english: bool = True,
    ) -> Optional[MusicInfo]:
        """
        Searches for a song by artist and title.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        limit: int, optional
            The number of items to retrieve from API. ``limit >=1 and <= 50``. Default is ``10``.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison. Default is ``True``.

        Returns
        -------
        Optional[MusicInfo_]
            The music information if found, otherwise None.
        """
        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        self._normalize_non_english = normalize_non_english

        music_info = None
        artist_ids = None
        queries = [
            f"?q=artist:{artist} track:{song}&type=track&limit={limit}",
            f"?q=artist:{artist} album:{song}&type=album&limit={limit}",
        ]

        for query in queries:
            if music_info:
                return music_info

            self._refresh_access_token()

            query_url = f"{self.__api_url}/search{query}"

            logger.info(
                f"Searching Spotify for `artist='{artist}'` and `song='{song}'`"
            )
            logger.debug(f"Query URL: {query_url}")

            try:
                response = self._session.get(
                    query_url, headers=self._authorization_header(), timeout=30
                )
                response.raise_for_status()
            except requests.RequestException as e:
                logger.warning(f"Failed to search for music: {e}")
                return None

            artist_ids = artist_ids if artist_ids else self._get_artists_ids(artist)
            music_info = self._find_music_info(
                artist, song, response.json(), artist_ids
            )

        return music_info

    def search_advanced(
        self,
        artist: str,
        song: str,
        isrc: str = None,
        upc: str = None,
        limit: int = 1,
        normalize_non_english: bool = True,
    ) -> Optional[MusicInfo]:
        """
        Searches for a song by artist, title, ISRC, or UPC.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        isrc : str, optional
            The ISRC of the track.
        upc : str, optional
            The UPC of the album.
        limit: int, optional
            The number of items to retrieve from API. ``limit >=1 and <= 50``. Default is ``1``.
        normalize_non_english : bool, optional
            Whether to normalize non-English characters for comparison. Default is ``True``.

        Returns
        -------
        Optional[MusicInfo_]
            The music information if found, otherwise None.
        """
        if not is_valid_string(artist) or not is_valid_string(song):
            raise InvalidValueException(
                "Artist and song names must be valid strings and can't be empty."
            )

        self._normalize_non_english = normalize_non_english
        self._refresh_access_token()

        if isrc:
            query = f"?q={artist} {song} isrc:{isrc}&type=track&limit={limit}"
        elif upc:
            query = f"?q={artist} {song} upc:{upc}&type=album&limit={limit}"
        else:
            raise InvalidValueException("ISRC or UPC must be provided.")

        query_url = f"{self.__api_url}/search{query}"
        try:
            response = self._session.get(
                query_url, headers=self._authorization_header(), timeout=30
            )
            response.raise_for_status()
        except requests.RequestException as e:
            raise logger.warning(f"Failed to search music with ISRC/UPC: {e}")
            return None

        artist_ids = self._get_artists_ids(artist)
        return self._find_music_info(artist, song, response.json(), artist_ids)

    def _get_artists_ids(self, artist: str) -> Union[list, None]:
        """
        Retrieves the IDs of the artists.

        Parameters
        ----------
        artist : str
            The name of the artist.

        Returns
        -------
        Union[list, None]
            A list of artist IDs or None if not found.
        """
        artist_ids = []
        for name in separate_artists(artist):
            query_url = f"{self.__api_url}/search?q={name}&type=artist&limit=5"
            try:
                response = self._session.get(
                    query_url, headers=self._authorization_header(), timeout=30
                )
                response.raise_for_status()
            except requests.RequestException as e:
                logger.warning(f"Network error during Spotify get artist ids: {e}")
                return None

            if response.status_code != 200:
                return None

            artist_ids.extend(
                artist["id"] for artist in response.json()["artists"]["items"]
            )
        return artist_ids

    def _find_music_info(
        self, artist: str, song: str, response_json: dict, artist_ids: list
    ) -> Optional[MusicInfo]:
        """
        Finds the music information from the search results.

        Parameters
        ----------
        artist : str
            The name of the artist.
        song : str
            The title of the song.
        response_json : dict
            The JSON response from the API.
        artist_ids : list
            A list of artist IDs.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        try:
            for track in response_json["tracks"]["items"]:
                music_info = self._find_track(song, artist, track, artist_ids)
                if music_info:
                    return music_info
        except KeyError:
            pass

        try:
            for album in response_json["albums"]["items"]:
                music_info = self._find_album(song, artist, album, artist_ids)
                if music_info:
                    return music_info
        except KeyError:
            pass

        logger.warning(
            f"No matching results found for artist='{artist}' and song='{song}'"
        )
        return None

    def _find_track(
        self, song: str, artist: str, track: dict, artist_ids: list
    ) -> Optional[MusicInfo]:
        """
        Finds the track information from the search results.

        Parameters
        ----------
        song : str
            The title of the song.
        artist : str
            The name of the artist.
        track : dict
            A single track from the search results.
        artist_ids : list
            A list of artist IDs.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        if not are_strings_similar(
            track["name"],
            song,
            use_translation=self._normalize_non_english,
            translation_session=self._translation_session,
        ):
            return None

        artists_name = [x["name"] for x in track["artists"]]
        matching_artists = [
            x["name"]
            for x in track["artists"]
            if are_strings_similar(
                x["name"],
                artist,
                use_translation=self._normalize_non_english,
                translation_session=self._translation_session,
            )
            or x["id"] in artist_ids
        ]

        if matching_artists:
            music_info = MusicInfo(
                album_art=track["album"]["images"][0]["url"],
                album_title=track["album"]["name"],
                album_type=track["album"]["album_type"],
                artists=", ".join(artists_name),
                genre=None,
                id=track["id"],
                isrc=track.get("external_ids").get("isrc"),
                lyrics=None,
                release_date=track["album"]["release_date"],
                tempo=None,
                title=track["name"],
                type="track",
                upc=None,
                url=track["external_urls"]["spotify"],
            )

            if self.fetch_lyrics:
                with LrcLib() as lrc_lib:
                    lyrics = lrc_lib.get_lyrics(
                        artist=music_info.artists, song=music_info.title
                    )
                if lyrics:
                    music_info.lyrics = lyrics.get("plainLyrics")
            return music_info
        return None

    def _find_album(
        self, song: str, artist: str, album: dict, artist_ids: list
    ) -> Optional[MusicInfo]:
        """
        Finds the album information from the search results.

        Parameters
        ----------
        song : str
            The title of the song.
        artist : str
            The name of the artist.
        album : dict
            A single album from the search results.
        artist_ids : list
            A list of artist IDs.

        Returns
        -------
        Optional[MusicInfo]
            The music information if found, otherwise None.
        """
        if not are_strings_similar(
            album["name"],
            song,
            use_translation=self._normalize_non_english,
            translation_session=self._translation_session,
        ):
            return None

        artists_name = [x["name"] for x in album["artists"]]
        matching_artists = [
            x["name"]
            for x in album["artists"]
            if are_strings_similar(
                x["name"],
                artist,
                use_translation=self._normalize_non_english,
                translation_session=self._translation_session,
            )
            or x["id"] in artist_ids
        ]

        if matching_artists:
            guess = guess_album_type(album.get("total_tracks", 1))
            guessed_right = are_strings_similar(
                album.get("album_type", "x"), guess, use_translation=False
            )

            return MusicInfo(
                album_art=album["images"][0]["url"],
                album_title=album["name"],
                album_type=album.get("album_type") if guessed_right else guess,
                artists=", ".join(artists_name),
                genre=None,
                id=album["id"],
                isrc=None,
                lyrics=None,
                release_date=album["release_date"],
                tempo=None,
                title=album["name"],
                type=album.get("type"),
                upc=None,
                url=album["external_urls"]["spotify"],
            )

        return None


class SpotifyAuth(BaseAuthClient):
    """
    A class to interact with the Spotify API. It uses "Authorization Code" grant type (or flow).

    This class reads the ``SPOTIFY_CLIENT_ID``, ``SPOTIFY_CLIENT_SECRET`` and ``SPOTIFY_REDIRECT_URI``
    from environment variables or the ``.env`` file by default.
    Alternatively, you can manually provide these values when creating an object.
    """

    def __init__(
        self,
        client_id: str = None,
        client_secret: str = None,
        redirect_uri: str = None,
        scopes: list[str] = None,
        defer_load: bool = False,
        fetch_lyrics: bool = True,
    ):
        """
        Parameters
        ----------
        client_id : str, optional
            The Client ID for the Spotify API. Defaults to ``SPOTIFY_CLIENT_ID`` from environment variable or the ``.env`` file.
        client_secret : str, optional
            The Client secret for the Spotify API. Defaults to ``SPOTIFY_CLIENT_SECRET`` from environment variable or the ``.env`` file.
        redirect_uri : str, optional
            The Redirect URI for the Spotify API. Defaults to ``SPOTIFY_REDIRECT_URI`` from environment variable or the ``.env`` file.
        scopes : list[str], optional
            A list of scopes for the Spotify API. For example: `['user-read-email', 'user-read-private']`.
        defer_load : bool, optional
            Whether to defer loading the access token during initialization. Default is ``False``.
        fetch_lyrics : bool, optional
            Whether to fetch lyrics using `LRCLIB <https://lrclib.net>`__ if the music platform does not provide lyrics (default is True).
        """
        self.client_id = client_id or os.getenv("SPOTIFY_CLIENT_ID")
        self.client_secret = client_secret or os.getenv("SPOTIFY_CLIENT_SECRET")
        self.redirect_uri = redirect_uri or os.getenv("SPOTIFY_REDIRECT_URI")
        self.scopes = scopes
        self.fetch_lyrics = fetch_lyrics

        if not self.client_id:
            raise SpotifyAuthException(
                "Client ID was not found. Set it in environment variable or directly pass it when creating object."
            )

        if not self.client_secret:
            raise SpotifyAuthException(
                "Client Secret was not found. Set it in environment variable or directly pass it when creating object."
            )

        if not self.redirect_uri:
            raise SpotifyAuthException(
                "No redirect URI was provided! Set it in environment variable or directly pass it when creating object."
            )

        if not scopes:
            logger.warning(
                "No scopes were provided. Authorization will only grant access to publicly available information."
            )
        else:
            self.scopes = " ".join(scopes)

        super().__init__(
            service_name="Spotify",
            access_token_url="https://accounts.spotify.com/api/token",
            user_auth_url="https://accounts.spotify.com/authorize",
            client_id=self.client_id,
            client_secret=self.client_secret,
            redirect_uri=self.redirect_uri,
            scopes=self.scopes,
            defer_load=defer_load,
        )

        self.__api_url = "https://api.spotify.com/v1/me"

    def get_user_profile(self) -> Optional[dict]:
        """
        Fetches the user's display name and profile images.

        Notes
        -----
        - ``user-read-email`` and ``user-read-private`` scopes are required to access user profile information.

        Returns
        -------
        dict
            A dictionary containing the user's display name and profile images.
        """
        self._refresh_access_token()
        query_url = self.__api_url
        header = self._authorization_header()

        try:
            response = self._session.get(query_url, headers=header, timeout=30)
            response.raise_for_status()
        except requests.RequestException as e:
            logger.warning(f"Failed to fetch user profile: {e}")
            return None

        if response.status_code != 200:
            logger.warning(f"Unexpected response: {response.json()}")
            return None

        result = response.json()
        return {
            "display_name": result.get("display_name"),
            "images": result.get("images", []),
            "url": result.get("external_urls", {}).get("spotify"),
        }

    def get_currently_playing(self) -> Optional[UserPlaying]:
        """
        Fetches information about the currently playing track for the authenticated user.

        This method interacts with the Spotify API to retrieve details about the track
        the user is currently listening to. It includes information such as the track's
        title, album, artists, release date, and more.

        Returns
        -------
        Optional[UserPlaying_]
            An instance of the ``UserPlaying`` model containing details about the currently
            playing track if available, or ``None`` if no track is currently playing or an
            error occurs.

        Notes
        -----
        - The user must have granted the necessary permissions (e.g., `user-read-currently-playing` scope) for this method to work.
        - If the API response does not contain the expected data, the method will return `None`.

        """
        query_url = f"{self.__api_url}/player/currently-playing"
        self._refresh_access_token()
        header = self._authorization_header()

        try:
            response = self._session.get(query_url, headers=header, timeout=30)
            response.raise_for_status()
        except requests.RequestException as e:
            logger.warning(f"Error while getting Spotify user activity: {e}")
            return None

        if response.status_code == 204:
            logger.info("Requested user is currently not listening to any music.")
            return None

        response_json = response.json()
        result = response_json.get("item")
        if result:
            guess = guess_album_type(result.get("album", {}).get("total_tracks", 1))
            guessed_right = are_strings_similar(
                result.get("album", {}).get("album_type", "x"),
                guess,
                use_translation=False,
            )
            # Spotify returns timestamp in milliseconds, so convert milliseconds to seconds:
            timestamp = response_json.get("timestamp") / 1000.0
            user_playing = UserPlaying(
                album_art=result.get("album", {}).get("images", [])[0].get("url"),
                album_title=result.get("album", {}).get("name"),
                album_type=(
                    result.get("album", {}).get("album_type")
                    if guessed_right
                    else guess
                ),
                artists=", ".join([x["name"] for x in result.get("artists", [])]),
                genre=None,
                id=result.get("id"),
                isrc=result.get("external_ids", {}).get("isrc"),
                is_playing=response_json.get("is_playing"),
                lyrics=None,
                release_date=result.get("album", {}).get("release_date"),
                tempo=None,
                timestamp=timestamp,
                title=result.get("name"),
                type=result.get("type"),
                upc=result.get("external_ids", {}).get("upc"),
                url=result.get("external_urls", {}).get("spotify"),
            )

            if self.fetch_lyrics:
                with LrcLib() as lrc_lib:
                    lyrics = lrc_lib.get_lyrics(
                        artist=user_playing.artists, song=user_playing.title
                    )
                if lyrics:
                    user_playing.lyrics = lyrics.get("plainLyrics")
            return user_playing
        return None


if __name__ == "__main__":
    import logging
    from dataclasses import asdict

    from yutipy.logger import enable_logging

    enable_logging(level=logging.DEBUG)

    print("\nChoose Spotify Grant Type/Flow:")
    print("1. Client Credentials (Spotify)")
    print("2. Authorization Code (SpotifyAuth)")
    choice = input("\nEnter your choice (1 or 2): ")

    if choice == "1":
        spotify = Spotify()

        try:
            artist_name = input("Artist Name: ")
            song_name = input("Song Name: ")
            result = spotify.search(artist_name, song_name)
            pprint(asdict(result))
        finally:
            spotify.close_session()

    elif choice == "2":
        redirect_uri = input("Enter Redirect URI: ")
        scopes = ["user-read-email", "user-read-private"]

        spotify_auth = SpotifyAuth(scopes=scopes)

        try:
            state = spotify_auth.generate_state()
            auth_url = spotify_auth.get_authorization_url(state=state)
            print(f"Opening the following URL in your browser: {auth_url}")
            webbrowser.open(auth_url)

            code = input("Enter the authorization code: ")
            spotify_auth.callback_handler(code, state, state)

            user_profile = spotify_auth.get_user_profile()
            if user_profile:
                print(f"Successfully authenticated \"{user_profile['display_name']}\".")
            else:
                print("Authentication successful, but failed to fetch user profile.")
        finally:
            spotify_auth.close_session()

    else:
        print("Invalid choice. Exiting.")



import pytest
from yutipy.lrclib import LrcLib
from yutipy.exceptions import InvalidValueException

class DummyResponse:
    def __init__(self, json_data, status_code=200):
        self._json = json_data
        self.status_code = status_code
    def json(self):
        return self._json
    def raise_for_status(self):
        if self.status_code != 200:
            raise Exception("HTTP error")

def test_init_default():
    lib = LrcLib()
    assert lib.api_url == "https://lrclib.net/api"
    assert lib.app_name == "yutipy"
    assert lib.app_url == "https://github.com/CheapNightbot/yutipy"
    assert not lib.is_session_closed


def test_init_custom():
    lib = LrcLib(app_name="foo", app_version="1.2.3", app_url="bar")
    assert lib.app_name == "foo"
    assert lib.app_version == "1.2.3"
    assert lib.app_url == "bar"


def test_get_lyrics_success(monkeypatch):
    # Simulate a successful API response with a matching result
    def mock_get(self, url, timeout):
        return DummyResponse([
            {"trackName": "Song", "artistName": "Artist", "plainLyrics": "Lyrics here"}
        ])
    monkeypatch.setattr("requests.Session.get", mock_get)
    lib = LrcLib()
    result = lib.get_lyrics("Artist", "Song")
    assert result is not None
    assert result["plainLyrics"] == "Lyrics here"


def test_get_lyrics_no_match(monkeypatch):
    def mock_get(self, url, timeout):
        return DummyResponse([])
    monkeypatch.setattr("requests.Session.get", mock_get)
    lib = LrcLib()
    result = lib.get_lyrics("Artist", "Song")
    assert result is None


def test_get_lyrics_invalid(monkeypatch):
    lib = LrcLib()
    with pytest.raises(InvalidValueException):
        lib.get_lyrics("", "Song")
    with pytest.raises(InvalidValueException):
        lib.get_lyrics("Artist", " ")


def test_context_manager(monkeypatch):
    lib = LrcLib()
    with lib as l:
        assert l is lib
    assert lib.is_session_closed


def test_close_session():
    lib = LrcLib()
    lib.close_session()
    assert lib.is_session_closed
    lib.close_session()
    assert lib.is_session_closed



import os
import webbrowser

from dotenv import load_dotenv, set_key


def run_config_wizard():
    """Interactive configuration wizard for setting up API keys."""
    print("Welcome to the yutipy Configuration Wizard!")
    print("This wizard will help you set up your API keys for various services.\n")

    # Load existing .env file if it exists
    env_file = ".env"
    load_dotenv(env_file)

    # List of available services and their required environment variables
    services = {
        "Spotify": {
            "SPOTIFY_CLIENT_ID": {
                "description": "Spotify Client ID",
                "url": "https://developer.spotify.com/dashboard",
                "instructions": """
1. Go to your Spotify Developer Dashboard: https://developer.spotify.com/dashboard
2. Create a new app and fill in the required details.
3. Copy the "Client ID" and "Client Secret" from the app's settings.
4. Paste them here when prompted.
                """,
            },
            "SPOTIFY_CLIENT_SECRET": {
                "description": "Spotify Client Secret",
                "url": "https://developer.spotify.com/dashboard",
                "instructions": "See the steps above for Spotify Client ID.",
            },
        },
        "KKBox": {
            "KKBOX_CLIENT_ID": {
                "description": "KKBox Client ID",
                "url": "https://developer.kkbox.com/",
                "instructions": """
1. Go to the KKBOX Developer Portal: https://developer.kkbox.com/
2. Log in and create a new application.
3. Copy the "Client ID" and "Client Secret" from the app's settings.
4. Paste them here when prompted.
                """,
            },
            "KKBOX_CLIENT_SECRET": {
                "description": "KKBox Client Secret",
                "url": "https://developer.kkbox.com/",
                "instructions": "See the steps above for KKBox Client ID.",
            },
        },
        "Last.fm": {
            "LASTFM_API_KEY": {
                "description": "Last.fm API Key",
                "url": "https://www.last.fm/api/account/create",
                "instructions": """
1. Go to the Last.fm API account creation page: https://www.last.fm/api/account/create
2. Log in with your Last.fm account.
3. Create a new application and fill in the required details.
4. Copy the "API Key" from the application settings.
5. Paste it here when prompted.
                """,
            },
        },
    }

    # Display available services
    print("Available services:")
    for i, service in enumerate(services.keys(), start=1):
        print(f"{i}. {service}")

    # Prompt the user to select a service
    choice = input("\nEnter the number of the service you want to configure: ").strip()
    try:
        service_name = list(services.keys())[int(choice) - 1]
    except (IndexError, ValueError):
        print("Invalid choice. Exiting configuration wizard.")
        return

    print(f"\nYou selected: {service_name}")

    # Get the selected service's variables
    selected_service = services[service_name]

    # Track whether the browser has already been opened for a service
    browser_opened = set()

    # Prompt the user for each variable in the selected service
    for var, details in selected_service.items():
        current_value = os.getenv(var)
        if current_value:
            print(f"{details['description']} is already set.")
            continue

        print(f"\n{details['description']} is missing.")
        print(details["instructions"])

        # Check if the browser has already been opened for this service
        if details["url"] not in browser_opened:
            open_browser = (
                input(
                    f"Do you want to open the website to get your {details['description']}? (y/N): "
                )
                .strip()
                .lower()
            )
            if open_browser == "y":
                webbrowser.open(details["url"])
                print(f"The website has been opened in your browser: {details['url']}")
                browser_opened.add(details["url"])  # Mark this URL as opened

        # Prompt the user to enter the value
        new_value = input(f"Enter your {details['description']}: ").strip()
        if new_value:
            set_key(env_file, var, new_value)
            print(f"{details['description']} has been saved to the .env file.")

    print("\nConfiguration complete! Your API keys have been saved to the .env file.")


if __name__ == "__main__":
    run_config_wizard()



# This file is intentionally left blank. /ᐠ｡ꞈ｡ᐟ\ 喵~ not anymore !


class BaseResponse:
    """
    A base mock response class for simulating HTTP responses in tests.

    Attributes:
        status_code (int): The HTTP status code of the response. Defaults to 200.

    Methods:
        raise_for_status(): Simulates the behavior of `requests.Response.raise_for_status()`
            by doing nothing, indicating a successful response with no exceptions raised.
    """

    status_code = 200

    @staticmethod
    def raise_for_status():
        """Simulates a successful response with no exceptions raised."""
        pass



from yutipy import (
    deezer,
    exceptions,
    itunes,
    kkbox,
    lastfm,
    logger,
    lrclib,
    musicyt,
    spotify,
    yutipy_music,
)

__all__ = [
    "deezer",
    "exceptions",
    "itunes",
    "kkbox",
    "lastfm",
    "logger",
    "lrclib",
    "musicyt",
    "spotify",
    "yutipy_music",
]



__all__ = ["LastFm", "LastFmException"]

import os
from pprint import pprint
from time import time
from typing import Optional

import requests
from dotenv import load_dotenv

from yutipy.exceptions import LastFmException
from yutipy.logger import logger
from yutipy.models import UserPlaying
from yutipy.utils.helpers import separate_artists

load_dotenv()

LASTFM_API_KEY = os.getenv("LASTFM_API_KEY")


class LastFm:
    """
    A class to interact with the Last.fm API for fetching user music data.

    This class reads the ``LASTFM_API_KEY`` from environment variables or the ``.env`` file by default.
    Alternatively, you can manually provide this values when creating an object.
    """

    def __init__(self, api_key: str = None):
        """
        Parameters
        ----------
        lastfm_api_key : str, optional
            The Lastfm API Key (<https://www.last.fm/api>). Defaults to ``LASTFM_API_KEY`` from environment variable or the ``.env`` file.

        Raises:
            LastFmException: If the API key is not provided or found in the environment.
        """
        self.api_key = api_key or LASTFM_API_KEY

        if not self.api_key:
            raise LastFmException(
                "Lastfm API key was not found. Set it in environment variable or directly pass it when creating object."
            )

        self._is_session_closed = False

        self.__api_url = "https://ws.audioscrobbler.com/2.0"
        self.__session = requests.Session()

    def __enter__(self):
        """Enters the runtime context related to this object."""
        return self

    def __exit__(self, exc_type, exc_value, exc_traceback):
        """Exits the runtime context related to this object."""
        self.close_session()

    def close_session(self) -> None:
        """Closes the current session(s)."""
        if not self._is_session_closed:
            self.__session.close()
            self._is_session_closed = True

    @property
    def is_session_closed(self) -> bool:
        """Checks if the session is closed."""
        return self._is_session_closed

    def get_user_profile(self, username: str):
        """
        Fetches the user profile information for the provided username.

        Returns
        -------
        dict
            A dictionary containing the user's profile information or error is username does not exist.
        """
        query = (
            f"?method=user.getinfo&user={username}&api_key={self.api_key}&format=json"
        )
        query_url = self.__api_url + query

        try:
            response = self.__session.get(query_url, timeout=30)
        except requests.RequestException as e:
            logger.warning(f"Failed to fetch user profile: {e}")
            return None

        response_json = response.json()
        result = response_json.get("user")
        error = response_json.get("message")
        if result:
            images = [
                {"size": image.get("size"), "url": image.get("#text")}
                for image in result.get("image", [])
            ]
            return {
                "name": result.get("realname"),
                "username": result.get("name"),
                "type": result.get("type"),
                "url": result.get("url"),
                "images": images,
            }
        elif error:
            return {"error": error}
        else:
            return None

    def get_currently_playing(self, username: str) -> Optional[UserPlaying]:
        """
        Fetches information about the currently playing or most recent track for a user.

        Parameters
        ----------
        username : str
            The Last.fm username to fetch data for.

        Returns
        -------
        Optional[UserPlaying_]
            An instance of the ``UserPlaying`` model containing details about the currently
            playing track if available, or ``None`` if the request fails or no data is available.
        """
        query = f"?method=user.getrecenttracks&user={username}&limit=1&api_key={self.api_key}&format=json"
        query_url = self.__api_url + query

        try:
            response = self.__session.get(query_url, timeout=30)
            response.raise_for_status()
        except requests.RequestException as e:
            logger.warning(f"Failed to fetch user profile: {e}")
            return None

        response_json = response.json()
        result = response_json.get("recenttracks", {}).get("track", [])[0]
        is_playing = result.get("@attr", {}).get("nowplaying", False)
        is_playing = (
            True if isinstance(is_playing, str) and is_playing == "true" else False
        )
        if result and is_playing:
            album_art = [
                img.get("#text")
                for img in result.get("image", [])
                if img.get("size") == "extralarge"
            ]
            return UserPlaying(
                album_art="".join(album_art),
                album_title=result.get("album", {}).get("#text"),
                artists=", ".join(
                    separate_artists(result.get("artist", {}).get("#text"))
                ),
                id=result.get("mbid") if result.get("mbid") else None,
                timestamp=result.get("date", {}).get("uts") or time(),
                title=result.get("name"),
                url=result.get("url"),
                is_playing=is_playing,
            )
        return None


if __name__ == "__main__":
    with LastFm() as lastfm:
        username = input("Enter Lasfm Username: ").strip()
        result = lastfm.get_user_profile(username=username)

        if result:
            pprint(result)
        else:
            print("No result was found. Make sure the username is correct!")
